#include "managers.h"
#include "global.h"
#include "xpms.h"
#include <algorithm>
#include <stdexcept>


template <class Ctr> inline void Delete(typename Ctr::value_type ptr)
	{ delete ptr; }


/*
 * WaveData
 */

int WaveData::WaveSize() const
{
	return X11::Inst().WindowWidth() * wave_size_pc /
	       g_aliens_hspacing          / 100;
}



/*
 *  BulletsManager
 */
 
BulletsManager::~BulletsManager()
{
	std::for_each(bullets_.begin(), bullets_.end(), Delete<BulletsCtr>);
}


void BulletsManager::Move()
{
	BulletsCtr::iterator it (bullets_.begin());
	while (it != bullets_.end()) {
		(*it)->Move();
		if ((*it)->Out()) it = Destroy(it);
		else ++it;
	}
}


void BulletsManager::Add(const Pix * pix, Coord pos, Coord speed)
{
	bullets_.push_back(new Bullet(pix, pos, speed));
}


int BulletsManager::DoCollisions(Coord dim, Coord pos, int max)
{
	int res (0);
	BulletsCtr::iterator it (bullets_.begin());
	while (res < max && it != bullets_.end()) {
		if ((*it)->Collide(pos, dim)) {
			++res;
			(*it)->Erase();
			it = Destroy(it);
		}
		else ++it;
	}
	return res;
}


BulletsManager::BulletsCtr::iterator
BulletsManager::Destroy(BulletsCtr::iterator it)
{
	delete *it;
	return bullets_.erase(it);
}



/*
 *  BonusManager
 */
 
BonusManager::~BonusManager()
{
	std::for_each(bonuses_.begin(), bonuses_.end(), Delete<BonusCtr>);
}


void BonusManager::Move()
{
	BonusCtr::iterator it (bonuses_.begin());
	while (it != bonuses_.end()) {
		(*it)->Move();
		if ((*it)->Out()) {
			delete *it;
			it = bonuses_.erase(it);
		}
		else ++it;
	}
}


void BonusManager::Add(const Pix * pix, Coord pos, Bonus::bonus_t type)
{
	bonuses_.push_back(new Bonus(pix, pos, type));
}


Bonus::bonus_t BonusManager::DoFirstCollision(Coord dim, Coord pos)
{
	BonusCtr::iterator it (bonuses_.begin());
	while (it != bonuses_.end()) {
		if ((*it)->Collide(pos, dim)) {
			const Bonus::bonus_t res ((*it)->Type());
			(*it)->Erase();
			delete *it;
			bonuses_.erase(it);
			return res;
		}
		++it;
	}
	return Bonus::none;
}



/*
 *  ExplosionsManager
 */

ExplosionsManager::~ExplosionsManager()
{
	std::for_each(explosions_.begin(), explosions_.end(), Delete<ExplosionsCtr>);
}


void ExplosionsManager::Move()
{
	ExplosionsCtr::iterator it (explosions_.begin());
	while (it != explosions_.end()) {
		(*it)->Move();
		if ((*it)->Finished()) {
			delete *it;
			it = explosions_.erase(it);
		}
		else ++it;
	}
}


void ExplosionsManager::Add(Coord pos, Coord speed, XColor color, int duration)
{
	explosions_.push_back(new Explosion(pos, speed, color, duration));
}



/*
 *  AliensManager
 */

int AliensManager::bonus_wait_;


AliensManager::AliensManager(BulletsManager * bombs_manager,
                             BulletsManager * bullets_manager,
                             BonusManager * bonus_manager,
			     ExplosionsManager * explosions_manager,
                             int level_number, const WaveData * waves_data,
                             int max_level)
: bombs_manager_     (bombs_manager)
, bullets_manager_   (bullets_manager)
, bonus_manager_     (bonus_manager)
, explosions_manager_(explosions_manager)
, level_number_      (level_number)
, waves_             (waves_data)
, nwaves_            (GetNWaves())
, max_wave_size_     (GetMaxWaveSize())
, wave_number_       (0)
, speed_             (level_number / (max_level + 1) + 3)
, base_cruise_       (Coord(0, 10))
, base_cruise_speed_ (1)
, next_creation_     (new_wave)
{
	if (level_number_ < 1 || !waves_ || !waves_[0].alien_xpm) {
		throw std::runtime_error("AliensManager data error");
	}
	next_creation_wait_ = waves_[0].wait;
	if (level_number_ == 1) bonus_wait_ = 20;
}


size_t AliensManager::GetNWaves() const
{
	const WaveData * p;
	for (p = waves_; p->alien_xpm; ++p);
	return p - waves_;
}


int AliensManager::GetMaxWaveSize() const
{
	int res (0);
	for (size_t i (0); i < nwaves_; ++i) {
		const int wave_size (waves_[i].WaveSize());
		if (wave_size > res) res = wave_size;
	}
	return res;
}


AliensManager::~AliensManager()
{
	std::for_each(aliens_.begin(), aliens_.end(), Delete<AliensCtr>);
}


void AliensManager::Move() {
	base_cruise_.x += base_cruise_speed_;
	if (base_cruise_speed_ < 0 && base_cruise_.x <= g_alien_width / 2 ||
	    base_cruise_speed_ > 0 && base_cruise_.x >=
	    max_wave_size_ * g_aliens_hspacing + g_alien_width / 2 +
	    (g_aliens_hspacing - g_alien_width)) {
		base_cruise_speed_ *= -1;
	}

	for (std::list<Alien*>::iterator it (aliens_.begin());
	     it != aliens_.end(); ++it) {
		(*it)->Move();
	}

	if (next_creation_wait_ == 0) Creation();
	else --next_creation_wait_;
}


void AliensManager::Fire() const
{
	for (AliensCtr::const_iterator it (aliens_.begin());
	     it != aliens_.end() &&
	     bombs_manager_->Nb() < 3u * level_number_ * max_wave_size_ / 10;
	     ++it) {
		if ((*it)->Stage() != Trajectory::cruising &&
		    (*it)->Stage() != Trajectory::joining &&
		    rand() % 50 < speed_) {
			bombs_manager_->Add((*it)->Fire());
		}
	}
}


int AliensManager::DoBulletsCollisions()
{
	int nb (0);
	AliensCtr::iterator ita (aliens_.begin());
	while (ita != aliens_.end()) {
	     	const Coord alien_pos ((*ita)->Position());
		if (bullets_manager_->DoCollisions((*ita)->Dim(), alien_pos, 1)) {
			// explosion
			(*ita)->Erase();
			XColor color;
			color.red = color.green = 65535; color.blue = 54*1024;
			explosions_manager_->Add(alien_pos, Coord(0,-4), color);
			// bonus
			CreateBonusMaybe((*ita)->Position());
			// alien destruction
			++nb;
			delete *ita;
			ita = aliens_.erase(ita);
		}
		else ++ita;
	}
	return nb;
}


void AliensManager::Creation()
{
	switch (next_creation_) {
	case new_wave:
		wave_alien_number_ = 0;
		next_creation_ = new_alien;
		// continue with new_alien
	case new_alien: {
		const WaveData *const wave (waves_ + wave_number_);
		const Pix *const pix (PixKeeper::Instance().Get(wave->alien_xpm));
		const Coord delta_cruise
			((wave_alien_number_ +
			  (max_wave_size_ - wave->WaveSize()) / 2) *
			 g_aliens_hspacing,
			 (nwaves_ - wave_number_) * g_aliens_vspacing);
		const bool mirrored
			(wave->mirrored && wave_alien_number_ % 2 == 0);
		const Trajectory trajectory
			(wave->arrival, mirrored, base_cruise_, delta_cruise);
		Alien *const alien (new Alien(pix, trajectory, speed_));
		aliens_.push_back(alien);

		if (++wave_alien_number_ < wave->WaveSize()) {
			next_creation_wait_ = 30 / speed_;
		} else {
			if (++wave_number_ < nwaves_) {
				next_creation_ = new_wave;
				next_creation_wait_ = waves_[wave_number_].wait;
			} else {
				next_creation_ = new_attack;
				next_creation_wait_ =
					std::max(50, 750 - level_number_ * 33);
			}
		}
		break;
	}
	case new_attack: {
		int nb_attacking (0);
		for (AliensCtr::const_iterator it (aliens_.begin());
		     it != aliens_.end(); ++it) {
			if ((*it)->Stage() == Trajectory::attacking) ++nb_attacking;
		}
		for (AliensCtr::iterator a (aliens_.begin());
		     a != aliens_.end() && nb_attacking < level_number_ * 3;
		     ++a) {
			if ((*a)->Stage() == Trajectory::cruising) {
				(*a)->BuildAttack();
				++nb_attacking;
			}
		}
		next_creation_wait_ = 5;
		break;
	}
	} // switch
}


void AliensManager::CreateBonusMaybe(Coord pos)
{
	if (--bonus_wait_ > 0) return;
	bonus_wait_ = max_wave_size_ * 5;
	switch (rand() % 5) {
	case 0:
		bonus_manager_->Add(PixKeeper::Instance().Get(bonus_speed_xpm),
		                    pos, Bonus::extra_speed);
		break;
	case 1: case 2: case 3:
		bonus_manager_->Add(PixKeeper::Instance().Get(bonus_fire_xpm),
		                    pos, Bonus::extra_fire);
		break;
	default:
		bonus_manager_->Add(PixKeeper::Instance().Get(bonus_shield_xpm),
		                    pos, Bonus::extra_shield);
	}
}


/*
 *  Player
 */

Player::Player(BulletsManager * bullets_manager,
               BulletsManager * bombs_manager,
	       BonusManager * bonus_manager,
               ExplosionsManager * explosions_manager,
               int below_y)
: Sprite             (PixKeeper::Instance().Get(player_xpm),
                      Coord(X11::Inst().WindowWidth() / 2, -50))
, bullets_manager_   (bullets_manager)
, bombs_manager_     (bombs_manager)
, bonus_manager_     (bonus_manager)
, explosions_manager_(explosions_manager)
, below_y_           (below_y)
, shield_            (3)
, speed_             (2)
, fire_wait_         (0)
, fire_interval_     (24)
, multi_fire_        (1)
{
}


void Player::Move(int direction)
{
	if (fire_wait_ > 0) --fire_wait_;

	Coord pos (Position());
	if (direction != 0) {
		Erase();
		pos.x += direction * speed_;
	}
	if (pos.x < Width() / 2) pos.x = Width() / 2;
	else {
		const int max_right (X11::Inst().WindowWidth() - Width() / 2 - 1);
		if (pos.x > max_right) pos.x = max_right;
	}
	pos.y = PosY();
	SetPos(pos);
	Draw();
}


void Player::Fire()
{
	if (fire_wait_ > 0) return;	
	fire_wait_ = fire_interval_;

	for (int i (0); i < multi_fire_; ++i) {
		Coord speed (i - multi_fire_ / 2, -10);
		if (multi_fire_ % 2 == 0 && i >= multi_fire_ / 2) ++speed.x;
		Coord fire_pos (Position());
		fire_pos.y -= Height() / 2;
		bullets_manager_->Add(PixKeeper::Instance().Get(bullet_xpm),
		                      fire_pos, speed);
	}
}


void Player::DoBombsCollisions()
{
	for (int n (bombs_manager_->DoCollisions(Dim(), Position())); n; --n) {
		--shield_;
		XColor color;
		color.blue = 65535; color.red = color.green = 54*1024;
		explosions_manager_->Add(Position(), Coord(0,0), color, 50);
	}
}


void Player::DoBonusCollisions()
{
	Bonus::bonus_t res;
	while ((res = bonus_manager_->DoFirstCollision(Dim(), Position())) != Bonus::none) {
		switch (res) {
		case Bonus::extra_speed:
			ExtraSpeed();
			break;
		case Bonus::extra_fire:
			ExtraFire();
			break;
		case Bonus::extra_shield:
			ExtraShield();
			break;
		default:
			break;
		}
	}
}


int Player::PosY() const
{
	return X11::Inst().WindowHeight() - below_y_ - Height() / 2;
}


void Player::ExtraFire()
{
	if (multi_fire_ < 9)
		if (fire_interval_ > 5) {
			fire_interval_ = fire_interval_ * 3 / 4;
			fire_wait_ = fire_wait_ * 3 / 4;
		} else {
			fire_interval_ = fire_interval_ *
			                 (multi_fire_+1) / multi_fire_;
			fire_wait_ = fire_wait_ *
			             (multi_fire_+1) / multi_fire_;
			++multi_fire_;
		}
}
