#ifndef MANAGERS_H
#define MANAGERS_H

#include "global.h"
#include "explosion.h"
#include "sprites.h"
#include <limits>
#include <list>
#include <utility>
#include <vector>


// An attack wave is defined by the alien vessel, the number of
// vessels and the attack formation (that can be X-mirrored)
struct WaveData {
	int wait;
	const char ** alien_xpm;
	int wave_size_pc;
	const RCoord * arrival;
	bool mirrored;
	
	int WaveSize() const;
};


class BulletsManager {
	typedef std::list<Bullet*> BulletsCtr;
	BulletsCtr bullets_;
public:
	BulletsManager() {}
	~BulletsManager();
	void Move();
	void Add(const Pix*, Coord pos, Coord speed);
	void Add(Bullet * b) { bullets_.push_back(b); }
	size_t Nb() const { return bullets_.size(); }
	int DoCollisions(Coord dim, Coord pos,
	                 int max = std::numeric_limits<int>::max());
private:
	BulletsCtr::iterator Destroy(BulletsCtr::iterator);
};


class BonusManager {
	typedef std::list<Bonus*> BonusCtr;
	BonusCtr bonuses_;
public:
	BonusManager() {}
	~BonusManager();
	void Move();
	void Add(const Pix * pix, Coord pos, Bonus::bonus_t type);
	Bonus::bonus_t DoFirstCollision(Coord dim, Coord pos);
};


class ExplosionsManager {
	typedef std::list<Explosion*> ExplosionsCtr;
	ExplosionsCtr explosions_;
public:
	ExplosionsManager() {}
	~ExplosionsManager();
	void Move();
	void Add(Coord pos, Coord speed, XColor color, int duration = 24);
};


// Aliens arrive in waves. The AlienManager creates aliens according
// to waves data and then move and destroy them if collision.
class AliensManager {
	static int bonus_wait_;

	BulletsManager *const bombs_manager_;
	BulletsManager *const bullets_manager_;
	BonusManager *const bonus_manager_;
	ExplosionsManager *const explosions_manager_;
	const int level_number_;
	const WaveData *const waves_;
	const size_t nwaves_;
	const int max_wave_size_;
	unsigned wave_number_;
	int wave_alien_number_;
	const int speed_;
	Coord base_cruise_;
	int base_cruise_speed_;
	enum creation_t { new_wave, new_alien, new_attack };
	creation_t next_creation_;
	int next_creation_wait_;
	typedef std::list<Alien*> AliensCtr;
	AliensCtr aliens_;
public:
	AliensManager(BulletsManager * bombs_manager,
	              BulletsManager * bullets_manager,
	              BonusManager*, ExplosionsManager*,
	              int level_number, const WaveData * wavesdata,
	              int max_level);
	~AliensManager();
	void Move();
	void Fire() const;
	int  DoBulletsCollisions();
	bool Finished() const
		{ return next_creation_ == new_attack && aliens_.empty(); }
private:
	size_t GetNWaves() const;
	int GetMaxWaveSize() const;
	void Creation();
	void CreateBonusMaybe(Coord);
};


// The player ship is displayed by a pix and moves on the bottom of the window.
// It can fire bullets. It can collide with alien bullets (bombs) and bonuses.
// The shield is decremented by each collision.
class Player : public Sprite {
	Player(const Player &); // no copy
	Player & operator=(const Player &); // no copy

	BulletsManager *const bullets_manager_;
	BulletsManager *const bombs_manager_;
	BonusManager *const bonus_manager_;
	ExplosionsManager *const explosions_manager_;
	const int below_y_;
	int shield_;
	int speed_;
	int fire_wait_, fire_interval_;
	int multi_fire_;
public:
	Player(BulletsManager * bullets_manager, BulletsManager * bombs_manager,
	       BonusManager*, ExplosionsManager*, int below_y);
	void Move(int direction);
	void Fire();
	void DoBombsCollisions();
	void DoBonusCollisions();
	int  Shield() const { return shield_; }
	void ExtraFire();
	void ExtraShield() { ++shield_; }
	void ExtraSpeed() { ++speed_; }
private:
	int  PosY() const;
};

#endif
